;; Generated package description from matlab-mode.el  -*- no-byte-compile: t -*-
(define-package "matlab-mode" "8.0.1.0.20260130.94305" "Major mode for MATLAB(R) dot-m files" '((emacs "27.2")) :commit "7144e1f4f2bffcda6cbe61bfca5e8c7cc5d2d76b" :authors '(("Matt Wette" . "mwette@alumni.caltech.edu") ("Eric M. Ludlam" . "eludlam@mathworks.com")) :maintainer '(("Eric M. Ludlam" . "eludlam@mathworks.com") ("Uwe Brauer" . "oub@mat.ucm.es") ("John Ciolfi" . "john.ciolfi.32@gmail.com")) :keywords '("matlab(r)") :url "https://github.com/mathworks/Emacs-MATLAB-Mode")
